#include <assert.h>
#include <errno.h>
#include <poll.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <unistd.h>
#include <string>

#include "acceptor.h"
#include "log.h"
#include "serverpool.h"
#include "state.pb.h"
#include "udpinput.h"
#include "version.h"

using namespace std;

extern ServerPool *servers;

UDPInput::UDPInput(const string &stream_id, const string &url)
	: stream_id(stream_id),
	  url(url),
	  sock(-1)
{
	// Should be verified by the caller.
	string protocol;
	bool ok = parse_url(url, &protocol, &host, &port, &path);
	assert(ok);

	construct_header();
}

UDPInput::UDPInput(const InputProto &serialized)
	: stream_id(serialized.stream_id()),
	  url(serialized.url()),
	  sock(serialized.sock())
{
	// Should be verified by the caller.
	string protocol;
	bool ok = parse_url(url, &protocol, &host, &port, &path);
	assert(ok);

	construct_header();
}

InputProto UDPInput::serialize() const
{
	InputProto serialized;
	serialized.set_stream_id(stream_id);
	serialized.set_url(url);
	serialized.set_sock(sock);
	return serialized;
}

void UDPInput::close_socket()
{
	int ret;
	do {
		ret = close(sock);
	} while (ret == -1 && errno == EINTR);

	if (ret == -1) {
		log_perror("close()");
	}

	sock = -1;
}
	
void UDPInput::construct_header()
{
	string header =
		"HTTP/1.0 200 OK\r\n"
		"Content-type: application/octet-stream\r\n"
		"Cache-control: no-cache\r\n"
		"Server: " SERVER_IDENTIFICATION "\r\n"
		"\r\n";
	servers->set_header(stream_id, header);
}

void UDPInput::do_work()
{
	while (!should_stop) {
		if (sock == -1) {
			int port_num = atoi(port.c_str());
			sock = create_server_socket(port_num, UDP_SOCKET);
			if (sock == -1) {
				log(WARNING, "[%s] UDP socket creation failed. Waiting 0.2 seconds and trying again...",
				             stream_id.c_str());
				usleep(200000);
				continue;
			}
		}

		// Since we are non-blocking, we need to wait for the right state first.
		// Wait up to 50 ms, then check should_stop.
		pollfd pfd;
		pfd.fd = sock;
		pfd.events = POLLIN;

		int nfds = poll(&pfd, 1, 50);
		if (nfds == 0 || (nfds == -1 && errno == EINTR)) {
			continue;
		}
		if (nfds == -1) {
			log_perror("poll");
			close_socket();
			continue;	
		}

		char buf[4096];
		int ret;
		do {
			ret = recv(sock, buf, sizeof(buf), 0);
		} while (ret == -1 && errno == EINTR);

		if (ret <= 0) {
			log_perror("recv");
			close_socket();
			continue;
		}
		
		servers->add_data(stream_id, buf, ret);
	}
}
